package cz.jzitnik.game.ui;

import cz.jzitnik.game.Game;
import cz.jzitnik.game.items.InventoryItem;
import cz.jzitnik.game.items.Item;
import cz.jzitnik.tui.ScreenRenderer;
import cz.jzitnik.tui.SpriteCombiner;
import cz.jzitnik.game.sprites.ui.Number;
import cz.jzitnik.tui.SpriteList;
import lombok.Getter;
import lombok.Setter;
import org.jline.terminal.MouseEvent;
import org.jline.terminal.Terminal;

import java.util.ArrayList;
import java.util.List;
import java.util.Optional;

@Getter
public class Inventory {
    private static final int INVENTORY_SIZE_PX = 470;
    private static int COLUMN_AMOUNT = 5;
    private static int ROW_AMOUNT = 4;

    private InventoryItem[] items = new InventoryItem[20];
    private InventoryItem[] hotbar = new InventoryItem[9];
    private SmallCraftingTable smallCraftingTable = new SmallCraftingTable( this);

    @Setter
    private int itemInhHandIndex = 0;
    @Setter
    private int selectedItemInv = -1;

    public Optional<Item> getItemInHand() {
        if (hotbar[itemInhHandIndex] == null) {
            return Optional.empty();
        }
        return Optional.of(hotbar[itemInhHandIndex].getItem());
    }

    public void decreaseItemInHand() {
        if (hotbar[itemInhHandIndex] == null) {
            return;
        }

        if (hotbar[itemInhHandIndex].getAmount() == 1) {
            hotbar[itemInhHandIndex] = null;
            return;
        }

        hotbar[itemInhHandIndex].setAmount(hotbar[itemInhHandIndex].getAmount() - 1);
    }

    private void placeItem(Item item) {
        var inventoryItem = new InventoryItem(item);

        // Try to place to hotbar
        for (int i = 0; i < hotbar.length; i++) {
            if (hotbar[i] == null) {
                hotbar[i] = inventoryItem;
                return;
            }
        }

        // Try to place to inventory
        for (int i = 0; i < items.length; i++) {
            if (items[i] == null) {
                items[i] = inventoryItem;
                break;
            }
        }

        // If inventory is full the item is lost
    }

    public void addItem(Item item) {
        if (!item.isStackable()) {
            placeItem(item);
            return;
        }

        // Try to stack in hotbar
        for (InventoryItem value : hotbar) {
            if (value != null && value.getItem().equals(item) && value.getAmount() < item.getStackAmount()) {
                value.setAmount(value.getAmount() + 1);
                return;
            }
        }

        // Try to stack in inventory
        for (InventoryItem inventoryItem : items) {
            if (inventoryItem != null && inventoryItem.getItem().equals(item) && inventoryItem.getAmount() < item.getStackAmount()) {
                inventoryItem.setAmount(inventoryItem.getAmount() + 1);
                return;
            }
        }

        // If stacking didn't work try to add it to inventory
        placeItem(item);
    }

    private Number.NumberState getNumberState(int digit) {
        return switch (digit) {
            case 0 -> Number.NumberState.ZERO;
            case 1 -> Number.NumberState.ONE;
            case 2 -> Number.NumberState.TWO;
            case 3 -> Number.NumberState.THREE;
            case 4 -> Number.NumberState.FOUR;
            case 5 -> Number.NumberState.FIVE;
            case 6 -> Number.NumberState.SIX;
            case 7 -> Number.NumberState.SEVEN;
            case 8 -> Number.NumberState.EIGHT;
            case 9 -> Number.NumberState.NINE;
            default -> throw new IllegalArgumentException("Unexpected number: " + digit);
        };
    }

    private String getNumberSprite(int number) {
        Number nm = new Number();
        if (number <= 9) {
            Number.NumberState numberState = getNumberState(number);

            return Number.fixForHotbar(nm.getSprite(numberState), 1);
        }

        String numStr = Integer.toString(number); // Convert to string
        char firstChar = numStr.charAt(0); // Get first digit
        char secondChar = numStr.charAt(1); // Get second digit

        int firstDigit = Character.getNumericValue(firstChar);
        int secondDigit = Character.getNumericValue(secondChar);

        Number.NumberState numberState1 = getNumberState(firstDigit);
        Number.NumberState numberState2 = getNumberState(secondDigit);

        return SpriteCombiner.combineTwoSprites( Number.fixForHotbar(nm.getSprite(numberState1), 2), Number.fixForHotbar(nm.getSprite(numberState2), 1));
    }

    private String getHotbarBackground() {
        StringBuilder sprite = new StringBuilder();
        for (int i = 0; i < 25; i++) {
            sprite.append("\033[38;5;245;48;5;245m▓".repeat(50));
            sprite.append("\n");
        }
        return sprite.toString();
    }

    public void renderHotbar(StringBuilder buffer, SpriteList spriteList, Terminal terminal, boolean isFull) {
        int termWidth = terminal.getWidth();
        int startLeft = (termWidth / 2) - (INVENTORY_SIZE_PX / 2);

        List<String> sprites = getSprites(hotbar, spriteList, (isFull ? selectedItemInv - 20 : itemInhHandIndex));

        for (int i = 0; i < 26; i++) {
            // Empty left space
            buffer.append("\033[0m ".repeat(Math.max(0, startLeft)));

            if (i == 0 || i == 25) {
                buffer.append("\033[38;5;231;48;5;231m▓".repeat(INVENTORY_SIZE_PX - 2));
            } else {
                for (int j = 0; j < 9; j++) {
                    buffer.append("\033[38;5;231;48;5;231m▓".repeat(2));
                    if (sprites.get(j) == null) {
                        buffer.append("\033[0m ".repeat(50));
                        continue;
                    }

                    String x = sprites.get(j).split("\n")[i];
                    buffer.append(x);
                }
            }
            buffer.append("\033[38;5;231;48;5;231m▓".repeat(2));
            buffer.append("\033[0m\n");
        }
    }

    public void renderFull(StringBuilder buffer, Terminal terminal, SpriteList spriteList) {
        int widthPixels = COLUMN_AMOUNT * (50 + 4) + 2;
        int heightPixels = ROW_AMOUNT * (25 + 1);

        int moveLeft = (terminal.getWidth() / 2) - (widthPixels / 2);
        int moveTop = (terminal.getHeight() / 2) - (heightPixels / 2);

        List<String> sprites = getSprites(items, spriteList, selectedItemInv);

        // Top center
        buffer.append("\n".repeat(Math.max(0, moveTop)));

        buffer.append("\033[0m ".repeat(moveLeft));
        buffer.append("\033[38;5;231;48;5;231m▓".repeat(widthPixels)).append("\033[0m\n");

        String[] craftingTable = smallCraftingTable.render(spriteList).toString().split("\n");
        int spacesFromTop = (ROW_AMOUNT - SmallCraftingTable.ROW_AMOUNT) / 2;

        for (int i = 0; i < ROW_AMOUNT; i++) {
            for (int j = 0; j < 26; j++) {
                buffer.append("\033[0m ".repeat(moveLeft));
                if (j < 25) {
                    buffer.append("\033[38;5;231;48;5;231m▓");
                    for (int k = 0; k < COLUMN_AMOUNT; k++) {
                        buffer.append("\033[38;5;231;48;5;231m▓".repeat(2));
                        var item = items[i * COLUMN_AMOUNT + k];
                        if (item == null) {
                            buffer.append("\033[0m ".repeat(50));
                            buffer.append("\033[38;5;231;48;5;231m▓".repeat(2));
                            continue;
                        }
                        var sprite = sprites.get(i * COLUMN_AMOUNT + k).split("\n");
                        buffer.append(sprite[j]);
                        buffer.append("\033[38;5;231;48;5;231m▓".repeat(2));
                    }
                    buffer.append("\033[38;5;231;48;5;231m▓");
                } else {
                    buffer.append("\033[38;5;231;48;5;231m▓".repeat(widthPixels));
                    if (i == 0) {
                        buffer.append("\033[0m").append(" ".repeat(20));
                        buffer.append("\033[38;5;231;48;5;231m▓".repeat(106));
                    }
                }

                if (i + 1 > spacesFromTop) {
                    int craftingIndex = ((i - spacesFromTop) * 26) + j;

                    if (craftingIndex >= 0 && craftingIndex < craftingTable.length) {
                        buffer.append("\033[0m").append(" ".repeat(20));
                        buffer.append(craftingTable[craftingIndex]);
                    }
                }

                buffer.append("\n");
            }
        }

        buffer.append("\n".repeat(10));

        renderHotbar(buffer, spriteList, terminal, true);
    }

    public void click(MouseEvent mouseEvent, Terminal terminal, ScreenRenderer screenRenderer, Game game) {
        if (mouseEvent.getType() != MouseEvent.Type.Pressed) {
            return;
        }

        int x = mouseEvent.getX();
        int y = mouseEvent.getY();

        int widthPixels = COLUMN_AMOUNT * (50 + 4) + 2;
        int heightPixels = ROW_AMOUNT * (25 + 1);

        int moveLeft = (terminal.getWidth() / 2) - (widthPixels / 2);
        int moveTop = (terminal.getHeight() / 2) - (heightPixels / 2);

        int fx = x - moveLeft;
        int fy = y - moveTop;

        int startLeftHotbar = (terminal.getWidth() / 2) - (INVENTORY_SIZE_PX / 2);
        int startTopHotbar = moveTop + heightPixels + 10;

        int startLeftCrafting = moveLeft + widthPixels + 20;
        int startTopCrafting = moveTop + 26;

        if (x >= startLeftCrafting && y >= startTopCrafting && y <= startTopCrafting + 52 && x <= startLeftCrafting + 106 + 54 + 7) {
            smallCraftingTable.click(mouseEvent.getX() - startLeftCrafting, mouseEvent.getY() - startTopCrafting);
            screenRenderer.render(game);
            return;
        }

        if (x >= startLeftHotbar && y >= startTopHotbar && x <= startLeftHotbar + INVENTORY_SIZE_PX && y <= startTopHotbar + 26) {
            int index = (x - startLeftHotbar) / 51;
            if (hotbar[index] == null && selectedItemInv != -1) {
                hotbar[index] = getSelectedItem();
            } else if (hotbar[index] != null) {
                selectedItemInv = 20 + index;
            }

            screenRenderer.render(game);
            return;
        }

        if (!(fx >= 0 && fx <= widthPixels && fy >=0 && fy <= heightPixels)) {
            return;
        }

        switch (mouseEvent.getButton()) {
            case Button1 -> {
                int blockX = fx / 54;
                int blockY = fy / 26;

                InventoryItem inventoryItem = items[blockY * COLUMN_AMOUNT + blockX];
                if (inventoryItem != null) {
                    selectedItemInv = blockY * COLUMN_AMOUNT + blockX;
                }
                if (inventoryItem == null && selectedItemInv != -1) {
                    items[blockY * COLUMN_AMOUNT + blockX] = getSelectedItem();
                }
            }
        }

        screenRenderer.render(game);
    }

    public List<String> getSprites(InventoryItem[] items, SpriteList spriteList, int selectedItem) {
        List<String> sprites = new ArrayList<>();

        for (int i = 0; i < items.length; i++) {
            var item = items[i];

            if (item == null) {
                if (i == selectedItem) {
                    sprites.add(getHotbarBackground());
                    continue;
                }

                sprites.add(null);
                continue;
            }

            if (i == selectedItem) {
                sprites.add(SpriteCombiner.combineTwoSprites(getHotbarBackground(), SpriteCombiner.combineTwoSprites(
                        spriteList.getSprite(item.getItem().getSprite()).getSprite(),
                        getNumberSprite(item.getAmount())
                )));
                continue;
            }

            sprites.add(SpriteCombiner.combineTwoSprites(
                    spriteList.getSprite(item.getItem().getSprite()).getSprite(),
                    getNumberSprite(item.getAmount())
            ));
        }

        return sprites;
    }

    public InventoryItem getSelectedItem() {
        InventoryItem temp;
        if (selectedItemInv < 20) {
            temp = items[selectedItemInv];
            items[selectedItemInv] = null;
        } else if (selectedItemInv >= 29) {
            int index = selectedItemInv - 29;
            temp = smallCraftingTable.getItems()[index];
            smallCraftingTable.getItems()[index] = null;
        } else {
            int index = selectedItemInv - 20;
            temp = hotbar[index];
            hotbar[index] = null;
        }

        selectedItemInv = -1;

        return temp;
    }
}
